/* This is a file evolved from Folding MicroEmacs 3.10.21a */
/* built by Mike Burrow and Andy Sturgess at Inmos.        */
/* Folding MicroEmacs has ben tested on Sun and PC only.   */
/* Itself based on MicroEmacs 3.10 by Daniel M Kawrence    */
/* and Dave G Conroy.                                      */

/* SCCS Info @(#)tcap.c	1.4 95/08/11 */

/*
 * tcap:   Unix V5, SUN OS, SCO XENIX, V7 and BS4.2 Termcap video driver for
 * MicroEMACS 3.10
 * 
 * 12-10-88 - Modifications made by Guy Turcotte to accomodate SunOS V4.0 and
 * Xenix V2.2.1 :
 * 
 * SunOS mods:
 * 
 * o p_seq field of TBIND struct augmented to 10 chars to take into account
 * longer definitions for keys (some Sun's keys definitions need at least 7
 * chars...) as such, the code in get1key has been modified to take care of
 * the longer p_seq string.
 * 
 * o tcapopen modified to take care of the tgetstr problem (returns NULL on
 * undefined keys instead of a valid string pointer...)
 * 
 * o The timout algorithm of get1key has been modified to take care of the
 * following select() function problem: if some chars are already in the
 * terminal buffer before select is called and no others char appears on the
 * terminal, it will timeout anyway... (maybe a feature of SunOs V4.0)
 * 
 * Xenix mods:
 * 
 * o The first two points indicated above are applicable for the Xenix OS
 * 
 * o With my current knowledge, I can't find a clean solution to the timeout
 * problem of the get1key function under Xenix. I modified the code to get
 * rid of the BSD code (via the #if directive) and use the Xenix nap() and
 * rdchk() functions to make a 1/30 second wait. Seems to work as long as
 * there is not to much of activity from other processes on the system. (The
 * link command of the makefile must be modified to link with the x
 * library... you must add the option -lx)
 * 
 * o The input.c file has been modified to not include the get1key function
 * defined there in the case of USG. The #if directive preceeding the get1key
 * definition has been modified from
 * 
 * #if (V7 == 0) && (BSD == 0)
 * 
 * to:
 * 
 * #if (V7 == 0) && (BSD == 0) && (USG == 0)
 * 
 * o The following lines define the new termcap entry for the ansi kind of
 * terminal: it permits the use of functions keys F1 .. F10 and keys
 * HOME,END,PgUp,PgDn on the IBM PC keyboard (the last 3 lines of the
 * definition have been added):
 * 
 * li|ansi|Ansi standard crt:\
 * :al=\E[L:am:bs:cd=\E[J:ce=\E[K:cl=\E[2J\E[H:cm=\E[%i%d;%dH:co#80:\
 * :dc=\E[P:dl=\E[M:do=\E[B:bt=\E[Z:ei=:ho=\E[H:ic=\E[@:im=:li#25:\
 * :nd=\E[C:pt:so=\E[7m:se=\E[m:us=\E[4m:ue=\E[m:up=\E[A:\
 * :kb=^h:ku=\E[A:kd=\E[B:kl=\E[D:kr=\E[C:eo:sf=\E[S:sr=\E[T:\
 * :GS=\E[12m:GE=\E[10m:GV=\63:GH=D:\
 * :GC=E:GL=\64:GR=C:RT=^J:G1=?:G2=Z:G3=@:G4=Y:GU=A:GD=B:\
 * :CW=\E[M:NU=\E[N:RF=\E[O:RC=\E[P:\ :WL=\E[S:WR=\E[T:CL=\E[U:CR=\E[V:\
 * :HM=\E[H:EN=\E[F:PU=\E[I:PD=\E[G:\
 * :k1=\E[M:k2=\E[N:k3=\E[O:k4=\E[P:k5=\E[Q:\
 * :k6=\E[R:k7=\E[S:k8=\E[T:k9=\E[U:k0=\E[V:\
 * :kh=\E[H:kH=\E[F:kA=\E[L:kN=\E[G:kP=\E[I:
 * 
 * o  Andy Sturges - Inmos : Fixed function key handling for SunOS V.4 so it
 * does not rely on select as much for catching function keys.
 * 
 */

#define term_def

#include <stdio.h>


#include        "estruct.h"
#include        "etype.h"
#include        "edef.h"
#include        "elang.h"

#if AIX | SUNOS | LINUX
#include <signal.h>
#endif

#if TERMCAP



#if     USG | HPUX
#include        <time.h>
#endif
#if     BSD | V7
#include        <sys/time.h>
#include        <sys/types.h>
#endif

#define MARGIN  8
#define SCRSIZ  64
#define NPAUSE  10		/* # times thru update to pause */
#define BEL     0x07
#define ESC     0x1B

/* Termcap Sequence definitions    */

typedef struct TBIND
{
  char            p_name[4];	/* sequence name */
  short           p_code;	/* resulting keycode of sequence */
  char            p_seq[10];	/* terminal escape sequence */
}               TBIND;

#ifdef OLDTABLE

TBIND           ttable[] = {
  "bt", SHFT | C_TRL | 'i', "",	/* backtab */
  "k1", SPEC | '1', "",		/* function key 1 */
  "k2", SPEC | '2', "",		/* function key 2 */
  "k3", SPEC | '3', "",		/* function key 3 */
  "k4", SPEC | '4', "",		/* function key 4 */
  "k5", SPEC | '5', "",		/* function key 5 */
  "k6", SPEC | '6', "",		/* function key 6 */
  "k7", SPEC | '7', "",		/* function key 7 */
  "k8", SPEC | '8', "",		/* function key 8 */
  "k9", SPEC | '9', "",		/* function key 9 */
  "k0", SPEC | '0', "",		/* function key 10 */
  "kA", C_TRL | 'O', "",		/* insert line */
  "kb", C_TRL | 'H', "",		/* backspace */
  "kC", C_TRL | 'L', "",		/* clear screen */
  "kD", SPEC | 'D', "",		/* delete character */
  "kd", SPEC | 'N', "",		/* down cursor */
  "kE", C_TRL | 'K', "",		/* clear to end of line */
  "kF", C_TRL | 'V', "",		/* scroll down */
  "kH", SPEC | '>', "",		/* home down [END?] key */
  "kh", SPEC | '<', "",		/* home */
  "kI", SPEC | 'C', "",		/* insert character */
  "kL", C_TRL | 'K', "",		/* delete line */
  "kl", SPEC | 'B', "",		/* left cursor */
  "kN", SPEC | 'V', "",		/* next page */
  "kP", SPEC | 'Z', "",		/* previous page */
  "kR", C_TRL | 'Z', "",		/* scroll down */
  "kr", SPEC | 'F', "",		/* right cursor */
  "ku", SPEC | 'P', "",		/* up cursor */
};

#else

TBIND           ttable[] = {
  "k1", SPEC | '1', "",		/* kf1 /* function key 1 */
  "k2", SPEC | '2', "",		/* kf2 /* function key 2 */
  "k3", SPEC | '3', "",		/* kf3 /* function key 3 */
  "k4", SPEC | '4', "",		/* kf4 /* function key 4 */
  "k5", SPEC | '5', "",		/* kf5 /* function key 5 */
  "k6", SPEC | '6', "",		/* kf6 /* function key 6 */
  "k7", SPEC | '7', "",		/* kf7 /* function key 7 */
  "k8", SPEC | '8', "",		/* kf8 /* function key 8 */
  "k9", SPEC | '9', "",		/* kf9 /* function key 9 */
  "k0", SPEC | '0', "",		/* kf0 /* function key 10 */
  "l1", SPEC | SHFT | '1', "",	/* kl1 /* shftd func key 1 */
  "l2", SPEC | SHFT | '2', "",	/* kl2 /* shftd func key 2 */
  "l3", SPEC | SHFT | '3', "",	/* kl3 /* shftd func key 3 */
  "l4", SPEC | SHFT | '4', "",	/* kl4 /* shftd func key 4 */
  "l5", SPEC | SHFT | '5', "",	/* kl5 /* shftd func key 5 */
  "l6", SPEC | SHFT | '6', "",	/* kl6 /* shftd func key 6 */
  "l7", SPEC | SHFT | '7', "",	/* kl7 /* shftd func key 7 */
  "l8", SPEC | SHFT | '8', "",	/* kl8 /* shftd func key 8 */
  "l9", SPEC | SHFT | '9', "",	/* kl9 /* shftd func key 9 */
  "l0", SPEC | SHFT | '0', "",	/* kl0 /* shftd func key 10 */
  /* the following key-pad keys and their bindings are obvious */
  "kh", SPEC | '<', "",		/* khome /* home */
  "ku", SPEC | 'P', "",		/* kcuu1 /* up cursor */
  "kP", SPEC | 'Z', "",		/* kpp /* previous page */
  "kl", SPEC | 'B', "",		/* kcub1 /* left cursor */
  "kr", SPEC | 'F', "",		/* kcuf1 /* right cursor */
  "kH", SPEC | '>', "",		/* kll /* home down [END] */
  "kd", SPEC | 'N', "",		/* kcud1 /* down cursor */
  "kN", SPEC | 'V', "",		/* knp /* next page */
  /* the following are less obvious */
  /* ins          # kI kich1  insert character *//* bound to insspace */
  "kI", SPEC | 'C', "",		/* kich1 /* insert character */
  /* shift del    # kD kdch1  delete character *//* bound to forwdel */
  "kD", SPEC | 'D', "",		/* kdch1 /* delete character */
  /* shift ins    # kA kil1  insert line *//* bound to openline */
  "kA", C_TRL | 'O', "",		/* kil1 /* insert line */
  /* shift up     # kR kri  scroll up *//* bound to backpage */
  "kR", C_TRL | 'Z', "",		/* kri /* scroll up */
  /* shift down   # kF kind  scroll down *//* bound to forwpage */
  "kF", C_TRL | 'V', "",		/* kind /* scroll down */
  /* shift left   # kC kclr  clear screen *//* bound to backword */
  /* "kC",   META | 'B',             "",     /* kclr /* clear screen */
  "kC", SPEC | C_TRL | 'B', "",	/* kclr /* clear screen */
  /* shift right  # kE kel  clear to eol *//* bound to forwword */
  /* "kE",   META | 'F',             "",     /* kel /* clear to eol */
  "kE", SPEC | C_TRL | 'F', "",	/* kel /* clear to eol */
  /* shift pgup   # K3 ka3  pad upper right *//* bound to gotobop */
  /* "K3",   META | 'P',             "",     /* ka3 /* pad upper right */
  "K3", SPEC | C_TRL | 'Z', "",	/* ka3 /* pad upper right */
  /* shift pgdn   # K5 kc3  pad lower right *//* bound to gotoeop */
  /* "K5",   META | 'N',             "",     /* kc3 /* pad lower right */
  "K5", SPEC | C_TRL | 'V', "",	/* kc3 /* pad lower right */
  /* shift home   # K1 ka1  pad upper left *//* unbound key */
  "K1", SPEC | C_TRL | '<', "",	/* ka1 /* pad upper left */
  /* shift end    # K4 kc1  pad lower left *//* unbound key */
  "K4", SPEC | C_TRL | '>', "",	/* kc1 /* pad lower left */
  /* shift tab    # bt cbt backtab                /* unbound key */
  "bt", SHFT | C_TRL | 'i', "",	/* cbt /* backtab */
  /* and let's forget the following */
  /* alt del      # kL kdl1  delete line */
  /* "kL",   C_TRL | 'K',             "",     /* kdl1 /* delete line */
  /* "kb",   C_TRL | 'H',             "",     /* kbs /* backspace */
  /* "kC",   C_TRL | 'L',             "",     /* kclr /* clear screen */
  /* "kE",   C_TRL | 'K',             "",     /* kel /* clear to eol */
};

#endif

#define NTBINDS sizeof(ttable)/sizeof(TBIND)

extern int      ttopen();
extern int      ttgetc();
extern int      ttputc();
extern int      tgetnum();
extern int      ttflush();
extern int      ttclose();
extern int      tcapkopen();
extern int      tcapkclose();
extern int      tcapgetc();
extern int      tcapmove();
extern int      tcapeeol();
extern int      tcapeeop();
extern int      tcapbeep();
extern int      tcaprev();
extern int      tcapcres();
extern int      tcapopen();
extern int      tcapclose();
extern int      tput();
extern char    *tgoto();
#if     COLOR
extern int      tcapfcol();
extern int      tcapbcol();
#endif

#define TCAPSLEN 1024
char            tcapbuf[TCAPSLEN];
char           *UP, PC, *CM, *CE, *CL, *SO, *SE, *IS, *KS, *KE;

TERM            term = {
  0, 0, 0, 0,			/* these four values are set dynamically at
				 * open time */
  MARGIN,
  SCRSIZ,
  NPAUSE,
  tcapopen,
  tcapclose,
  tcapkopen,
  tcapkclose,
  tcapgetc,
  ttputc,
  ttflush,
  tcapmove,
  tcapeeol,
  tcapeeop,
  tcapbeep,
  tcaprev,
  tcapcres
#if     COLOR
  ,tcapfcol,
  tcapbcol
#endif
};

/* input buffers and pointers      */

#define IBUFSIZE        64	/* this must be a power of 2 */

unsigned char   in_buf[IBUFSIZE];	/* input character buffer */
int             in_next = 0;	/* pos to retrieve next input character */
int             in_last = 0;	/* pos to place most recent input character */

in_init()			/* initialize the input buffer */

{
  in_next = in_last = 0;
}

in_check()			/* is the input buffer non-empty? */

{
  if (in_next == in_last)
    return (FALSE);
  else
    return (TRUE);
}

in_put(event)

  int             event;	/* event to enter into the input buffer */

{
  in_buf[in_last++] = event;
  in_last &= (IBUFSIZE - 1);
}

int
in_get()			/* get an event from the input buffer */

{
  register int    event;	/* event to return */

  event = in_buf[in_next++];
  in_next &= (IBUFSIZE - 1);
  return (event);
}



static char     tcbuf[1024];
static char     err_str[72];
static char    *tv_stype;

char           *tgetstr();
char           *getenv();

#if AIX | SUNOS | LINUX

int win_busy = 1;
int win_flag = 0;

void win_update() {
  int row, col;
#if AIX
  if ((row=atoi(termdef(0,'l'))) > 0)
    newsize(TRUE, row);
  if ((col=atoi(termdef(0,'c') )) > 0)
    newwidth(TRUE, col);
#endif
#if SUNOS | LINUX
  if ((tgetent(tcbuf, tv_stype)) != 1)
    return;
  if ((row = (short) tgetnum("li")) > 0)
    newsize(TRUE, row);
  if ((col = (short) tgetnum("co")) > 0)
    newwidth(TRUE, col);
#endif
  win_flag = 0;
  update(TRUE);
}

void           *
win_handler(sig, code, scp, addr)
  int             sig, code;
  struct sigcontext *scp;
  char           *addr;
{
  if(sig == SIGWINCH)
    if(win_busy)
      win_flag = 1;
    else
      win_update();
#if AIX
  signal(SIGWINCH, win_handler);
#endif
}
#endif


/*
 * Open the terminal put it in RA mode learn about the screen size read
 * TERMCAP strings for function keys
 */

tcapopen()

{
  register int    index;	/* general index */
  char           *t, *p;
  int             i;

  if ((tv_stype = getenv("TERM")) == NULL)
  {
    puts(TEXT182);
    /* "Environment variable TERM not defined!" */
    meexit(1);
  }
  if ((tgetent(tcbuf, tv_stype)) != 1)
  {
    sprintf(err_str, TEXT183, tv_stype);
    /* "Unknown terminal type %s!" */
    puts(err_str);
    meexit(1);
  }
  term.t_mrow = 100;		/* term.t_nrow; */
  term.t_mcol = 200;		/* term.t_ncol; */
#if AIX
  if ((term.t_nrow=(short)atoi(termdef(0,'l') ))<=0)
#endif
  {
    if ((term.t_nrow = (short) tgetnum("li")) == -1)
    {
      puts(TEXT184);
      /* "termcap entry incomplete (lines)" */
      meexit(1);
    }
  }
#if AIX
    if ((term.t_ncol=(short)atoi(termdef(0,'c')))<=0)
#endif
    {
    if ((term.t_ncol = (short) tgetnum("co")) == -1)
    {
      puts(TEXT185);
      /* "Termcap entry incomplete (columns)" */
      meexit(1);
    }
  }
  term.t_nrow --;
  p = tcapbuf;
  t = tgetstr("pc", &p);
  if (t)
    PC = *t;

  CL = tgetstr("cl", &p);
  CM = tgetstr("cm", &p);
  CE = tgetstr("ce", &p);
  UP = tgetstr("up", &p);
  SE = tgetstr("se", &p);
  SO = tgetstr("so", &p);
  if (SO != NULL)
    revexist = TRUE;

  if (CL == NULL || CM == NULL || UP == NULL)
  {
    puts(TEXT186);
    /* "Incomplete termcap entry\n" */
    meexit(1);
  }
  if (CE == NULL)		/* will we be able to use clear to EOL? */
    eolexist = FALSE;

  IS = tgetstr("is", &p);	/* extract init string */
  KS = tgetstr("ks", &p);	/* extract keypad transmit string */
  KE = tgetstr("ke", &p);	/* extract keypad transmit end string */

  /* read definitions of various function keys into ttable */
  for (index = 0; index < NTBINDS; index++)
  {
    strcpy(ttable[index].p_seq,
	   fixnull(tgetstr(ttable[index].p_name, &p)));
  }

  /* tell unix we are goint to use the terminal */
  ttopen();

  /* make sure we don't over run the buffer (TOO LATE I THINK) */
  if (p >= &tcapbuf[TCAPSLEN])
  {
    puts(TEXT187);
    /* "Terminal description too big!\n" */
    meexit(1);
  }
  /* send init strings if defined */
  if (IS != NULL)
    putpad(IS);
#if MOUSE
  if (strncmp(tv_stype,"xterm",5)==0) 
    putpad("\033[?1000h");
#endif
  if (KS != NULL)
    putpad(KS);

  /* initialize the input buffer */
  in_init();
#if AIX | SUNOS | LINUX
  signal(SIGWINCH, win_handler);
#endif
}

tcapclose()
{
  /* send end-of-keypad-transmit string if defined */
  if (KE != NULL)
    putpad(KE);
#if MOUSE
  if (strncmp(tv_stype,"xterm",5)==0) 
    putpad("\033[?1000l");
#endif
  ttclose();
#if AIX | SUNOS | LINUX
  signal(SIGWINCH, SIG_DFL);
#endif
}

tcapkopen()

{
  strcpy(sres, "NORMAL");
}

tcapkclose()

{
}

unsigned int
extcode(c)

  unsigned int    c;

{
  return (c);
}

/*
 * TCAPGETC:       Get on character.  Resolve and setup all the appropriate
 * keystroke escapes as defined in the comments at the beginning of input.c
 */

int
tcapgetc()

{
  int             c;		/* current extended keystroke */
  /* if there are already keys waiting.... send them */
  if (in_check())
    return (in_get());

  /* otherwise... get the char for now */
  c = get1key();
#if MOUSE
  if (c & MOUS) {
    in_put(0);
    in_put(MOUS>>8);
    in_put(get1key());  /* col */
    in_put(get1key());  /* row */
    in_put(c & (~MOUS));
    return(tcapgetc());
  }
#endif
  /* unfold the control bit back into the character */
  if (C_TRL & c)
    c = (c & ~C_TRL) - '@';

  /* fold the event type into the input stream as an escape seq */
  if ((c & ~255) != 0)
  {
    in_put(0);			/* keyboard escape prefix */
    in_put(c >> 8);		/* event type */
    in_put(c & 255);		/* event code */
    return (tcapgetc());
  }
  return (c);
}

/*
 * GET1KEY:        Get one keystroke. The only prefixs legal here are the
 * SPEC and C_TRL prefixes.
 * 
 * Note:
 * 
 * Escape sequences that are generated by terminal function and cursor keys
 * could be confused with the user typing the default META prefix followed by
 * other chars... ie
 * 
 * UPARROW  =  <ESC>A   on some terminals... apropos  =  M-A
 * 
 * The difference is determined by measuring the time between the input of the
 * first and second character... if an <ESC> is types, and is not followed by
 * another char in 1/30 of a second (think 300 baud) then it is a user input,
 * otherwise it was generated by an escape sequence and should be SPECed.
 */

 /* The DAN version ignores time delay, because it causes problems over a
  * network.  We accept that some old types of terminal might cause
  * ambiguity.  We rely on all multicharacter sequencies starting with
  * ESCAPE and NOT containing ESCAPE.
  */

#define MAX_CSEQ 10
static int down_mouse = 0;  
int PASCAL      NEAR
get1key()

{
  int    c;
  int match;
  int    index;	/* index into termcap binding table */
  static int first;
  static int last = 0;
  static char cseq[MAX_CSEQ];
  if (last==0) 
    c = ttgetc();
  else {
    c = cseq[first];
    first++;
    if (first>=last) 
      last = 0;
    return(c);
  }
  if (c!= 27)
    return(c);
#if KBDTIME
  /* process a possible escape sequence */
  /* set up to check the keyboard for input */

#if     BSD | V7 | HPUX
  FD_ZERO(&fdset);
  FD_SET(0, &fdset);
  timeout.tv_sec = 0;
#if SUNOS
  timeout.tv_usec = 400000L;
#else
  timeout.tv_usec = 35000L;
#endif
  /* check to see if things are pending soon */
  if (kbdmode != PLAY &&
      select(1, &fdset, (int *) NULL, (int *) NULL, &timeout) == 0)
    return (C_TRL | '[');
#endif

#if XENIX
  if ((kbdmode != PLAY) && (rdchk(0) <= 0))
  {
    nap(35000L);
    if (rdchk(0) <= 0)
      return (C_TRL | '[');
  }
#endif

#if     USG
  /*
   * we don't know how to do this check for a pending char within 1/30th of a
   * second machine independantly in the general System V case.... so we
   * don't
   */
  if (kbdmode != PLAY)
    return (C_TRL | '[');
#endif

  /* a key is pending within 1/30 of a sec... its an escape sequence */
#endif
  cseq[0] = c;
  last = 1;
  match = 1;
  while(last<MAX_CSEQ && match) {
    match = 0;
    for (index = 0; index < NTBINDS; index++) {
      if (strncmp(cseq, ttable[index].p_seq, last) == 0) {
        match = 1;
        if (ttable[index].p_seq[last] == '\0') {
          last = 0;
          return ttable[index].p_code;
        }
        else {
          cseq[last] = ttgetc();
          last++;	
        }
      }
    }
    /*  Support for an XTERM mouse */
    if (strncmp(cseq,"\033[M",last) == 0) {
      if (last == 3) {
      	c = ttgetc() & 3;
        cseq[0] = ttgetc() - 33; /* col */
        cseq[1] = ttgetc() - 33; /* row */
        first = 0; last = 2;
        if ( c == 3) {
          c = (down_mouse&4)?2:((down_mouse&2)?1:0);
          down_mouse &= ~(1 << c);          
          return (MOUS | ('a' + 2*c + 1));
        }
        else {
          down_mouse |= 1 << c;
          return(MOUS | ('a' + 2*c));
        }
      }
      else {
        cseq[last] = ttgetc();
        last++;
        match = 1;	
      }
    }
  }
  first = 1;
  return(cseq[0]);
} 

tcapmove(row, col)
  register int    row, col;
{
  putpad(tgoto(CM, col, row));
}

tcapeeol()
{
  putpad(CE);
}

tcapeeop()
{
  putpad(CL);
}

tcaprev(state)			/* change reverse video status */

  int             state;	/* FALSE = normal video, TRUE = reverse video */

{
  /* static int revstate = FALSE; */

  if (state)
  {
    if (SO != NULL)
      putpad(SO);
  } else if (SE != NULL)
    putpad(SE);
}

tcapcres()			/* change screen resolution */

{
  return (TRUE);
}

spal(dummy)			/* change palette string */

{
  /* Does nothing here       */
}

#if     COLOR
tcapfcol()			/* no colors here, ignore this */
{
}

tcapbcol()			/* no colors here, ignore this */
{
}
#endif

tcapbeep()
{
  ttputc(BEL);
}

putpad(str)
  char           *str;
{
  tputs(str, 1, ttputc);
}

putnpad(str, n)
  char           *str;
{
  tputs(str, n, ttputc);
}


#if     FLABEL
fnclabel(f, n)			/* label a function key */

  int             f, n;		/* default flag, numeric argument [unused] */

{
  /* on machines with no function keys...don't bother */
  return (TRUE);
}
#endif
#else

hello()
{
}

#endif
