/*	EVAR.H: Environment and user variable definitions
		for MicroEMACS

		written 1993 by Daniel Lawrence
*/

/*	structure to hold user variables and their definitions	*/

typedef struct UVAR {
	char u_name[NVSIZE + 1];	       /* name of user variable */
	char *u_value;				/* value (string) */
} UVAR;

/*	current user variables (This structure will probably change)	*/

#define MAXVARS 	512

UVAR NOSHARE uv[MAXVARS + 1];	/* user variables */

/*	list of recognized environment variables	*/

NOSHARE char *envars[] = {
	"acount",		/* # of chars until next auto-save */
	"asave",		/* # of chars between auto-saves */
	"bufhook",		/* enter buffer switch hook */
	"cbflags",		/* current buffer flags */
	"cbufname",		/* current buffer name */
	"cfname",		/* current file name */
	"cmdhook",		/* command loop hook */
	"cmode",		/* mode of current buffer */
	"curchar",		/* current character under the cursor */
	"curcol",		/* current column pos of cursor */
	"curline",		/* current line in file */
	"curwidth",		/* current screen width */
	"curwind",		/* current window ordinal on current screen */
	"cwline",		/* current screen line in window */
	"debug",		/* macro debugging */
	"deskcolor",		/* desktop color */
	"diagflag",		/* diagonal mouse movements enabled? */
	"discmd",		/* display commands on command line */
	"disinp",		/* display command line input characters */
	"disphigh",		/* display high bit characters escaped */
	"exbhook",		/* exit buffer switch hook */
	"fcol",			/* first displayed column in curent window */
	"fillcol",		/* current fill column */
	"flicker",		/* flicker supression */
	"fmtlead",		/* format command lead characters */
        "foldafter",            /* after fold header mark DAN */
        "foldend",              /* end of fold mark DAN */
        "foldstart",            /* start of fold mark DAN */
	"gflags",		/* global internal emacs flags */
	"gmode",		/* global modes */
	"hardtab",		/* current hard tab size */
	"hilight",		/* region # to hilight (255 to turn off) */
	"hjump",		/* horizontal screen jump size */
	"hscrlbar",		/* horizontal scroll bar flag */
	"hscroll",		/* horizontal scrolling flag */
	"isterm",		/* incremental-search terminator character */
	"kill", 		/* kill buffer (read only) */
	"language",		/* language of text messages */
	"lastkey",		/* last keyboard char struck */
	"lastmesg",		/* last string mlwrite()ed */
	"line", 		/* text of current line */
        "lmargin",              /* left margin (enclosing fold) */
	"lterm",		/* current line terminator for writes */
        "ltype",                /* line type, normal, start of fold etc */
	"lwidth",		/* width of current line */
	"match",		/* last matched magic pattern */
	"mmove",		/* mouse moves events style */
	"modeflag",		/* Modelines displayed flag */
	"msflag",		/* activate mouse? */
	"numwind",		/* number of windows on current screen */
	"orgcol",		/* screen origin column */
	"orgrow",		/* screen origin row */
	"os",			/* what Operating System? */
	"overlap",		/* line overlap when paging */
	"pagelen",		/* number of lines used by editor */
	"palette",		/* current palette string */
	"paralead",		/* paragraph leadin characters */
	"pending",		/* type ahead pending flag */
	"popflag",		/* pop-up windows active? */
	"posflag",		/* display point position on modeline? */
	"progname",		/* returns current prog name - "MicroEMACS" */
	"ram",			/* ram in use by malloc */
	"readhook",		/* read file execution hook */
	"region",		/* current region (read only) */
	"replace",		/* replacement pattern */
	"rval", 		/* child process return value */
	"scrname",		/* current screen name */
	"search",		/* search pattern */
	"searchpnt",		/* differing search styles (term point) */
	"seed", 		/* current random number seed */
	"softtab",		/* current soft tab size */
	"sres", 		/* current screen resolution */
	"ssave",		/* safe save flag */
	"sscroll",		/* smooth scrolling flag */
	"status",		/* returns the status of the last command */
	"sterm",		/* search terminator character */
	"target",		/* target for line moves */
	"time",			/* date and time */
	"timeflag",		/* display time? */
	"tpause",		/* length to pause for paren matching */
	"version",		/* current version number */
	"vscrlbar",		/* vertical scroll bar flag */
	"wchars",		/* set of characters legal in words */
	"wline",		/* # of lines in current window */
	"wraphook",		/* wrap word execution hook */
	"writehook",		/* write file hook */
	"xpos", 		/* current mouse X position */
	"yankflag",		/* point placement at yanked/included text */
	"ypos",	 		/* current mouse Y position */
        "zpos"                  /* current entered fold depth */
};        

#define NEVARS	sizeof(envars) / sizeof(char *)

/*	and its preprocesor definitions 	*/

#define EVACOUNT	0
#define EVASAVE 	1
#define	EVBUFHOOK	2
#define EVCBFLAGS	3
#define EVCBUFNAME	4
#define EVCFNAME	5
#define EVCMDHK 	6
#define EVCMODE 	7
#define EVCURCHAR	8
#define EVCURCOL	9
#define EVCURLINE	10
#define EVCURWIDTH	11
#define	EVCURWIND	12
#define EVCWLINE	13
#define EVDEBUG         14
#define	EVDESKCLR	15
#define EVDIAGFLAG      16
#define EVDISCMD        17
#define EVDISINP        18
#define	EVDISPHIGH	19
#define EVEXBHOOK       20
#define EVFCOL		21
#define EVFILLCOL	22
#define EVFLICKER	23
#define	EVFMTLEAD	24
#define EVFOLDAFTER     25
#define EVFOLDEND       26
#define EVFOLDSTART     27
#define EVGFLAGS	28
#define EVGMODE 	29
#define	EVHARDTAB	30
#define	EVHILITE	31
#define EVHJUMP		32
#define EVHSCRLBAR      33
#define EVHSCROLL	34
#define EVISTERM	35
#define EVKILL          36
#define EVLANG          37
#define EVLASTKEY       38
#define EVLASTMESG      39
#define EVLINE          40
#define EVLMARGIN       41
#define	EVLTERM		42
#define EVLTYPE         43
#define EVLWIDTH        44
#define EVMATCH         45
#define EVMMOVE		46
#define EVMODEFLAG      47
#define EVMSFLAG        48
#define	EVNUMWIND	49
#define	EVORGCOL	50
#define	EVORGROW	51
#define	EVOS		52
#define	EVOVERLAP	53
#define EVPAGELEN       54
#define EVPALETTE       55
#define	EVPARALEAD	56
#define EVPENDING       57
#define	EVPOPFLAG	58
#define	EVPOSFLAG	59
#define EVPROGNAME      60
#define EVRAM           61
#define EVREADHK        62
#define	EVREGION	63
#define EVREPLACE       64
#define EVRVAL          65
#define EVSCRNAME	66
#define EVSEARCH        67
#define EVSEARCHPNT	68
#define EVSEED          69
#define EVSOFTTAB	70
#define EVSRES          71
#define EVSSAVE         72
#define EVSSCROLL	73
#define EVSTATUS	74
#define EVSTERM 	75
#define EVTARGET	76
#define EVTIME		77
#define EVTIMEFLAG	78
#define EVTPAUSE	79
#define EVVERSION	80
#define EVVSCRLBAR      81
#define	EVWCHARS	82
#define EVWLINE 	83
#define EVWRAPHK	84
#define	EVWRITEHK	85
#define EVXPOS		86
#define	EVYANKFLAG	87
#define EVYPOS		88
#define EVZPOS          89

/*	list of recognized user functions	*/

typedef struct UFUNC {
	char *f_name;  /* name of function */
	int f_type;	/* 1 = monamic, 2 = dynamic */
} UFUNC;

#define NILNAMIC	0
#define MONAMIC 	1
#define DYNAMIC 	2
#define TRINAMIC	3

NOSHARE UFUNC funcs[] = {
	"abs", MONAMIC, 	/* absolute value of a number */
	"add", DYNAMIC,		/* add two numbers together */
	"and", DYNAMIC, 	/* logical and */
	"asc", MONAMIC, 	/* char to integer conversion */
	"ban", DYNAMIC, 	/* bitwise and	 9-10-87  jwm */
	"bin", MONAMIC, 	/* loopup what function name is bound to a key */
	"bno", MONAMIC, 	/* bitwise not */
	"bor", DYNAMIC, 	/* bitwise or	 9-10-87  jwm */
	"bxo", DYNAMIC, 	/* bitwise xor	 9-10-87  jwm */
	"cat", DYNAMIC, 	/* concatinate string */
	"chr", MONAMIC, 	/* integer to char conversion */
	"div", DYNAMIC, 	/* division */
	"env", MONAMIC, 	/* retrieve a system environment var */
	"equ", DYNAMIC, 	/* logical equality check */
	"exi", MONAMIC, 	/* check if a file exists */
	"fin", MONAMIC, 	/* look for a file on the path... */
	"gre", DYNAMIC, 	/* logical greater than */
	"gro", MONAMIC,		/* return group match in MAGIC mode */
	"gtc", NILNAMIC,	/* get 1 emacs command */
	"gtk", NILNAMIC,	/* get 1 charater */
	"ind", MONAMIC, 	/* evaluate indirect value */
	"isn", MONAMIC,		/* is the arg a number? */
	"lef", DYNAMIC, 	/* left string(string, len) */
	"len", MONAMIC, 	/* string length */
	"les", DYNAMIC, 	/* logical less than */
	"low", MONAMIC, 	/* lower case string */
	"mid", TRINAMIC,	/* mid string(string, pos, len) */
	"mod", DYNAMIC, 	/* mod */
	"neg", MONAMIC, 	/* negate */
	"not", MONAMIC, 	/* logical not */
	"or",  DYNAMIC, 	/* logical or */
	"rev", MONAMIC,		/* reverse */
	"rig", DYNAMIC, 	/* right string(string, pos) */
	"rnd", MONAMIC, 	/* get a random number */
	"seq", DYNAMIC, 	/* string logical equality check */
	"sgr", DYNAMIC, 	/* string logical greater than */
	"sin", DYNAMIC, 	/* find the index of one string in another */
	"sle", DYNAMIC, 	/* string logical less than */
	"slo", DYNAMIC,		/* set lower to upper char translation */
	"sub", DYNAMIC, 	/* subtraction */
	"sup", DYNAMIC,		/* set upper to lower char translation */
	"tim", DYNAMIC, 	/* multiplication */
	"tri", MONAMIC,		/* trim whitespace off the end of a string */
	"tru", MONAMIC, 	/* Truth of the universe logical test */
	"upp", MONAMIC, 	/* uppercase string */
	"xla", TRINAMIC		/* XLATE character string translation */
};

#define NFUNCS	sizeof(funcs) / sizeof(UFUNC)

/*	and its preprocesor definitions 	*/

#define UFABS		0
#define UFADD		1
#define UFAND		2
#define UFASCII 	3
#define UFBAND		4
#define UFBIND		5
#define UFBNOT		6
#define UFBOR		7
#define UFBXOR		8
#define UFCAT		9
#define UFCHR		10
#define UFDIV		11
#define UFENV		12
#define UFEQUAL 	13
#define UFEXIST 	14
#define UFFIND		15
#define UFGREATER	16
#define UFGROUP		17
#define UFGTCMD 	18
#define UFGTKEY 	19
#define UFIND		20
#define	UFISNUM		21
#define UFLEFT		22
#define UFLENGTH	23
#define UFLESS		24
#define UFLOWER 	25
#define UFMID		26
#define UFMOD		27
#define UFNEG		28
#define UFNOT		29
#define UFOR		30
#define UFREVERSE	31
#define UFRIGHT 	32
#define UFRND		33
#define UFSEQUAL	34
#define UFSGREAT	35
#define UFSINDEX	36
#define UFSLESS 	37
#define	UFSLOWER	38
#define UFSUB		39
#define	UFSUPPER	40
#define UFTIMES 	41
#define	UFTRIM		42
#define UFTRUTH 	43
#define UFUPPER 	44
#define UFXLATE 	45

